"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EvaluationRecord = exports.Evaluation = void 0;
const css_select_1 = __importDefault(require("css-select"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const crypto_1 = require("crypto");
const util_1 = require("@qualweb/util");
const evaluationRecord_object_1 = __importDefault(require("./evaluationRecord.object"));
exports.EvaluationRecord = evaluationRecord_object_1.default;
const endpoint = 'http://194.117.20.242/validate/';
class Evaluation {
    async getEvaluator(page, sourceHtml, stylesheets, url) {
        const [plainHtml, pageTitle, elements, browserUserAgent] = await Promise.all([
            page.evaluate(() => {
                return document.documentElement.outerHTML;
            }),
            page.title(),
            page.$$('*'),
            page.browser().userAgent()
        ]);
        let urlStructure = undefined;
        if (url) {
            urlStructure = this.parseUrl(url, page.url());
        }
        const processedHtml = {
            html: {
                plain: plainHtml
            },
            title: pageTitle,
            elementCount: elements.length
        };
        const viewport = page.viewport();
        const evaluator = {
            name: 'QualWeb',
            description: 'QualWeb is an automatic accessibility evaluator for webpages.',
            version: '3.0.0',
            homepage: 'http://www.qualweb.di.fc.ul.pt/',
            date: new Date().toISOString().replace(/T/, ' ').replace(/\..+/, ''),
            hash: crypto_1.randomBytes(40).toString('hex'),
            url: urlStructure,
            page: {
                viewport: {
                    mobile: !!viewport.isMobile,
                    landscape: !!viewport.isLandscape,
                    userAgent: browserUserAgent,
                    resolution: {
                        width: viewport.width,
                        height: viewport.height
                    }
                },
                dom: {
                    source: sourceHtml,
                    processed: processedHtml,
                    stylesheets
                }
            }
        };
        return evaluator;
    }
    async addQWPage(page) {
        await page.addScriptTag({
            path: require.resolve('@qualweb/qw-page').replace('index.js', 'qwPage.js')
        });
    }
    async executeACT(page, sourceHtml, stylesheets, options) {
        await page.addScriptTag({
            path: require.resolve('@qualweb/act-rules')
        });
        const metaElements = css_select_1.default('meta', sourceHtml.html.parsed);
        const parsedMetaElements = new Array();
        for (let element of metaElements || []) {
            const content = util_1.DomUtils.getSourceElementAttribute(element, 'content');
            const httpEquiv = util_1.DomUtils.getSourceElementAttribute(element, 'http-equiv');
            const htmlCode = util_1.DomUtils.getSourceElementHtmlCode(element, true, false);
            const selector = util_1.DomUtils.getSourceElementSelector(element);
            parsedMetaElements.push({
                content,
                httpEquiv,
                htmlCode,
                selector
            });
        }
        const actReport = await page.evaluate((parsedMetaElements, stylesheets, options) => {
            const act = new ACTRules.ACTRules(options);
            return act.execute(parsedMetaElements, new QWPage.QWPage(document, window), stylesheets);
        }, parsedMetaElements, stylesheets, options);
        const r40 = 'QW-ACT-R40';
        if (!options || !options['rules'] || options['rules'].includes(r40) || options['rules'].includes('59br37')) {
            const viewport = page.viewport();
            await page.setViewport({
                width: 640,
                height: 512
            });
            const actReportR40 = await page.evaluate(() => {
                const act = new ACTRules.ACTRules();
                return act.executeQW_ACT_R40(new QWPage.QWPage(document, window));
            });
            await page.setViewport({
                width: viewport.width,
                height: viewport.height
            });
            actReport.assertions[r40] = actReportR40;
            actReport.metadata.passed += actReportR40.metadata.passed;
            actReport.metadata.failed += actReportR40.metadata.failed;
            actReport.metadata.warning += actReportR40.metadata.warning;
            if (actReportR40['outcome'] === 'inapplicable') {
                actReport.metadata.inapplicable++;
            }
        }
        return actReport;
    }
    async executeHTML(page, options) {
        await page.addScriptTag({
            path: require.resolve('@qualweb/html-techniques')
        });
        const url = page.url();
        const urlVal = await page.evaluate(() => {
            return location.href;
        });
        const validationUrl = endpoint + encodeURIComponent(urlVal);
        let response = undefined;
        let validation;
        try {
            response = await node_fetch_1.default(validationUrl);
        }
        catch (err) {
            console.error(err);
        }
        if (response && response.status === 200) {
            validation = JSON.parse(await response.json());
        }
        const newTabWasOpen = await util_1.BrowserUtils.detectIfUnwantedTabWasOpened(page.browser(), url);
        const htmlReport = await page.evaluate((newTabWasOpen, validation, options) => {
            const html = new HTMLTechniques.HTMLTechniques(options);
            return html.execute(new QWPage.QWPage(document, window), newTabWasOpen, validation);
        }, newTabWasOpen, validation, options);
        return htmlReport;
    }
    async executeCSS(page, stylesheets, mappedDOM, options) {
        await page.addScriptTag({
            path: require.resolve('@qualweb/css-techniques')
        });
        const cssReport = await page.evaluate((stylesheets, mappedDOM, options) => {
            const css = new CSSTechniques.CSSTechniques(options);
            return css.execute(new QWPage.QWPage(document, window), stylesheets, mappedDOM);
        }, stylesheets, {}, options);
        return cssReport;
    }
    async executeBP(page, options) {
        await page.addScriptTag({
            path: require.resolve('@qualweb/best-practices')
        });
        const bpReport = await page.evaluate((options) => {
            const bp = new BestPractices.BestPractices();
            if (options)
                bp.configure(options);
            const report = bp.execute(new QWPage.QWPage(document, window));
            return report;
        }, options);
        return bpReport;
    }
    async evaluatePage(sourceHtml, page, stylesheets, mappedDOM, execute, options, url) {
        const evaluator = await this.getEvaluator(page, sourceHtml, stylesheets, url);
        const evaluation = new evaluationRecord_object_1.default(evaluator);
        const reports = new Array();
        await this.addQWPage(page);
        if (execute.act) {
            reports.push(await this.executeACT(page, sourceHtml, stylesheets, options['act-rules']));
        }
        if (execute.html) {
            reports.push(await this.executeHTML(page, options['html-techniques']));
        }
        if (execute.css) {
            reports.push(await this.executeCSS(page, stylesheets, mappedDOM, options['css-techniques']));
        }
        if (execute.bp) {
            reports.push(await this.executeBP(page, options['best-practices']));
        }
        for (const report of reports || []) {
            if (report.type === 'wappalyzer') {
                evaluation.addModuleEvaluation('wappalyzer', report);
            }
            else if (report.type === 'act-rules') {
                evaluation.addModuleEvaluation('act-rules', report);
            }
            else if (report.type === 'html-techniques') {
                evaluation.addModuleEvaluation('html-techniques', report);
            }
            else if (report.type === 'css-techniques') {
                evaluation.addModuleEvaluation('css-techniques', report);
            }
            else if (report.type === 'best-practices') {
                evaluation.addModuleEvaluation('best-practices', report);
            }
        }
        return evaluation;
    }
    parseUrl(url, pageUrl) {
        let inputUrl = url;
        let protocol;
        let domainName;
        let domain;
        let uri;
        let completeUrl = pageUrl;
        protocol = completeUrl.split('://')[0];
        domainName = completeUrl.split('/')[2];
        const tmp = domainName.split('.');
        domain = tmp[tmp.length - 1];
        uri = completeUrl.split('.' + domain)[1];
        const parsedUrl = {
            inputUrl,
            protocol,
            domainName,
            domain,
            uri,
            completeUrl
        };
        return parsedUrl;
    }
}
exports.Evaluation = Evaluation;
