import CSSselect from 'css-select';
import {
  Page
} from 'puppeteer';
import {
  QualwebOptions,
  SourceHtml,
  CSSStylesheet,
  ProcessedHtml,
  Url,
  Evaluator
} from '@qualweb/core';
import fetch, { Response } from 'node-fetch';
import { HTMLValidationReport } from '@qualweb/html-validator';
import {
  randomBytes
} from 'crypto';
import {
  CSSTOptions,
  CSSTechniquesReport
} from '@qualweb/css-techniques';
import {
  BrowserUtils,
  DomUtils
} from '@qualweb/util';
import EvaluationRecord from './evaluationRecord.object';
import {
  ACTROptions, ACTRulesReport
} from '@qualweb/act-rules';
import {
  BPOptions, BestPracticesReport
} from '@qualweb/best-practices';
import {
  HTMLTOptions, HTMLTechniquesReport
} from '@qualweb/html-techniques';

const endpoint = 'http://194.117.20.242/validate/';

class Evaluation {

  public async getEvaluator(page: Page, sourceHtml: SourceHtml, stylesheets: CSSStylesheet[], url: string): Promise<Evaluator> {
    const [plainHtml, pageTitle, elements, browserUserAgent] = await Promise.all([
      page.evaluate(() => {
        return document.documentElement.outerHTML;
      }),
      page.title(),
      page.$$('*'),
      page.browser().userAgent()
    ]);

    let urlStructure: Url | undefined = undefined;
    if (url) {
      urlStructure = this.parseUrl(url, page.url())
    }

    const processedHtml: ProcessedHtml = {
      html: {
        plain: plainHtml
      },
      title: pageTitle,
      elementCount: elements.length
    };

    const viewport = page.viewport();

    const evaluator = {
      name: 'QualWeb',
      description: 'QualWeb is an automatic accessibility evaluator for webpages.',
      version: '3.0.0',
      homepage: 'http://www.qualweb.di.fc.ul.pt/',
      date: new Date().toISOString().replace(/T/, ' ').replace(/\..+/, ''),
      hash: randomBytes(40).toString('hex'),
      url: urlStructure,
      page: {
        viewport: {
          mobile: !!viewport.isMobile,
          landscape: !!viewport.isLandscape,
          userAgent: browserUserAgent,
          resolution: {
            width: viewport.width,
            height: viewport.height
          }
        },
        dom: {
          source: sourceHtml,
          processed: processedHtml,
          stylesheets
        }
      }
    };

    return evaluator;
  }

  public async addQWPage(page: Page): Promise<void> {
    await page.addScriptTag({
      path: require.resolve('@qualweb/qw-page').replace('index.js', 'qwPage.js')
    });
  }

  public async executeACT(page: Page, sourceHtml: SourceHtml, stylesheets: CSSStylesheet[], options: ACTROptions | undefined): Promise<ACTRulesReport> {
    await page.addScriptTag({
      path: require.resolve('@qualweb/act-rules')
    });

    const metaElements = CSSselect('meta', sourceHtml.html.parsed);
    const parsedMetaElements = new Array<any>();

    for (let element of metaElements || []) {
      const content = DomUtils.getSourceElementAttribute(element, 'content');
      const httpEquiv = DomUtils.getSourceElementAttribute(element, 'http-equiv');
      const htmlCode = DomUtils.getSourceElementHtmlCode(element, true, false);
      const selector = DomUtils.getSourceElementSelector(element);

      parsedMetaElements.push({
        content,
        httpEquiv,
        htmlCode,
        selector
      });
    }

    const actReport = await page.evaluate((parsedMetaElements, stylesheets, options) => {
      // @ts-ignore
      const act = new ACTRules.ACTRules(options);
      // @ts-ignore
      return act.execute(parsedMetaElements, new QWPage.QWPage(document, window), stylesheets);
      // @ts-ignore
    }, parsedMetaElements, stylesheets, options);

    const r40 = 'QW-ACT-R40';
    
    if (!options || !options['rules'] || options['rules'].includes(r40) || options['rules'].includes('59br37')) {
      const viewport = page.viewport();
      
      await page.setViewport({
        width: 640,
        height: 512
      });

      const actReportR40 = await page.evaluate(() => {
        // @ts-ignore
        const act = new ACTRules.ACTRules();
        // @ts-ignore
        return act.executeQW_ACT_R40(new QWPage.QWPage(document, window));
      });

      await page.setViewport({
        width: viewport.width,
        height: viewport.height
      });

      actReport.assertions[r40] = actReportR40;
      actReport.metadata.passed += actReportR40.metadata.passed;
      actReport.metadata.failed += actReportR40.metadata.failed;
      actReport.metadata.warning += actReportR40.metadata.warning;

      if (actReportR40['outcome'] === 'inapplicable') {
        actReport.metadata.inapplicable++;
      }
    }

    return actReport;
  }

  public async executeHTML(page: Page, options: HTMLTOptions | undefined): Promise<HTMLTechniquesReport> {
    await page.addScriptTag({
      path: require.resolve('@qualweb/html-techniques')
    });

    const url = page.url();
    const urlVal = await page.evaluate(() => {
      return location.href;
    });

    const validationUrl = endpoint + encodeURIComponent(urlVal);

    let response: Response | undefined = undefined;
    let validation: HTMLValidationReport;

    try {
      response = await fetch(validationUrl);
    } catch (err) {
      console.error(err);
    }

    if (response && response.status === 200) {
      validation = <HTMLValidationReport> JSON.parse(await response.json());
    }

    const newTabWasOpen = await BrowserUtils.detectIfUnwantedTabWasOpened(page.browser(), url);

    const htmlReport = await page.evaluate((newTabWasOpen, validation, options) => {
      // @ts-ignore
      const html = new HTMLTechniques.HTMLTechniques(options);
      // @ts-ignore
      return html.execute(new QWPage.QWPage(document, window), newTabWasOpen, validation);
      // @ts-ignore
    }, newTabWasOpen, validation, options);

    return htmlReport;
  }

  public async executeCSS(page: Page, stylesheets: CSSStylesheet[], mappedDOM: any, options: CSSTOptions | undefined): Promise<CSSTechniquesReport> {
    await page.addScriptTag({
      path: require.resolve('@qualweb/css-techniques')
    });

    const cssReport = await page.evaluate((stylesheets, mappedDOM, options) => {
      // @ts-ignore
      const css = new CSSTechniques.CSSTechniques(options);
      // @ts-ignore
      return css.execute(new QWPage.QWPage(document, window), stylesheets, mappedDOM);
      // @ts-ignore
    }, stylesheets, {}, options);

    return cssReport;
  }

  public async executeBP(page: Page, options: BPOptions): Promise<BestPracticesReport> {
    await page.addScriptTag({
      path: require.resolve('@qualweb/best-practices')
    });

    const bpReport = await page.evaluate((options) => {
      // @ts-ignore
      const bp = new BestPractices.BestPractices();
      if (options)
        bp.configure(options)
      // @ts-ignore
      const report = bp.execute(new QWPage.QWPage(document, window));
      return report;
      // @ts-ignore
    }, options);
    return bpReport;
  }

  public async evaluatePage(sourceHtml: SourceHtml, page: Page, stylesheets: CSSStylesheet[], mappedDOM: any, execute: any, options: QualwebOptions, url: string): Promise<EvaluationRecord> {
    const evaluator = await this.getEvaluator(page, sourceHtml, stylesheets, url);
    const evaluation = new EvaluationRecord(evaluator);

    const reports = new Array<any>();
    await this.addQWPage(page);

    if (execute.act) {
      reports.push(await this.executeACT(page, sourceHtml, stylesheets, options['act-rules']));
    }
    if (execute.html) {
      reports.push(await this.executeHTML(page, options['html-techniques']));
    }
    if (execute.css) {
      reports.push(await this.executeCSS(page, stylesheets, mappedDOM, options['css-techniques']));
    }
    if (execute.bp) {
      reports.push(await this.executeBP(page, options['best-practices']));
    }

    for (const report of reports || []) {
      if (report.type === 'wappalyzer') {
        evaluation.addModuleEvaluation('wappalyzer', report);
      } else if (report.type === 'act-rules') {
        evaluation.addModuleEvaluation('act-rules', report);
      } else if (report.type === 'html-techniques') {
        evaluation.addModuleEvaluation('html-techniques', report);
      } else if (report.type === 'css-techniques') {
        evaluation.addModuleEvaluation('css-techniques', report);
      } else if (report.type === 'best-practices') {
        evaluation.addModuleEvaluation('best-practices', report);
      }
    }

    return evaluation;
  }

  private parseUrl(url: string, pageUrl: string): Url {
    let inputUrl: string = url;
    let protocol: string;
    let domainName: string;
    let domain: string;
    let uri: string;
    let completeUrl: string = pageUrl;

    protocol = completeUrl.split('://')[0];
    domainName = completeUrl.split('/')[2];

    const tmp: string[] = domainName.split('.');
    domain = tmp[tmp.length - 1];
    uri = completeUrl.split('.' + domain)[1];

    const parsedUrl = {
      inputUrl,
      protocol,
      domainName,
      domain,
      uri,
      completeUrl
    };

    return parsedUrl;
  }
}

export {
  Evaluation,
  EvaluationRecord
};